#include "fxlib.h"
#include "Gray.h"
#include "timer.h"
#include "Syscall.h"

void Gray_Init() {
	static char buffer_DD_Dark[1024] = { 0 };
	static char buffer_DD_Light[1024] = { 0 };
	static char buffer_VRAM_Dark[1024] = { 0 };
	static char buffer_VRAM_Light[1024] = { 0 };
	Gray.DD.Dark = buffer_DD_Dark;
	Gray.DD.Light = buffer_DD_Light;
	Gray.VRAM.Dark = buffer_VRAM_Dark;
	Gray.VRAM.Light = buffer_VRAM_Light;
	System_GetOSVersion(Gray.OS);
	Gray.Initialised = TRUE;
}

// -1 = Error : 1 = Started
int Gray_Start() {
	if (Gray.Initialised != TRUE)
		Gray_Init();
	DisableGetkeyToMainFunctionReturn();
	return SetTimer(ID_USER_TIMER5, 25, Gray_Display_GrayScale) == -1 ? -1 : 1;
}

//-2 = Error : 0 = Stopped
int Gray_Stop() {
	EnableGetkeyToMainFunctionReturn();
	return KillTimer(ID_USER_TIMER5) == -2 ? -2 : 0;
}

//-1 = Error : 0 = Stopped : 1 = Started
int Gray_Toggle() {
	if (Gray_Stop() == 0)
		return 0;
	else
		return Gray_Start();
}

// 1 = Display Logo : 0 = Don't Display Logo
void Gray_PowerOffCalc(Bool logo) {
	unsigned int key;
	Bool Gray = Gray_Stop();
	if (logo == FALSE)
		while (GetKeyWait(KEYWAIT_HALTOFF_TIMEROFF, 0, !FALSE, &key));
	PowerOff(logo);
	if (!Gray == TRUE)
		Gray_Start();
}

void Gray_Copy_Buffer(char fromBuffer[1024], char toBuffer[1024]) {
	int i;
	for (i = 0; i < 1024; i++)
		toBuffer[i] = fromBuffer[i];
}

void Gray_Swap_Buffers() {
	char* tempBufferDark = Gray.DD.Dark;
	char* tempBufferLight = Gray.DD.Light;
	Gray.DD.Dark = Gray.VRAM.Dark;
	Gray.DD.Light = Gray.VRAM.Light;
	Gray.VRAM.Dark = tempBufferDark;
	Gray.VRAM.Light = tempBufferLight;
}

void Gray_Display_DD(unsigned char* buffer, int row, int LCD_reg_sel_1, int LCD_data_reg, register int LCD_reg_sel_2) {
	unsigned char* LCD_register_selector = (char*)0xB4000000;
	unsigned char* LCD_data_register = (char*)0xB4010000;
	int i;
	for (i = 0; i < 64; i++) {
		*LCD_register_selector = LCD_reg_sel_1;
		*LCD_data_register = row++;
		*LCD_register_selector = LCD_reg_sel_1;
		*LCD_data_register = LCD_data_reg;
		*LCD_register_selector = LCD_reg_sel_2;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
		*LCD_data_register = *buffer++;
	}
}

void Gray_Display_GrayScale() {
	if (Gray.OS[1] == '3')
		Gray_Display_DD(Gray.DD.Light, 0x80, 8, 4, 10);
	else
		Gray_Display_DD(Gray.DD.Light, 0xC0, 4, 0, 7);

	Gray_Delay_ASM(Gray.Time);

	if (Gray.OS[1] == '3')
		Gray_Display_DD(Gray.DD.Dark, 0x80, 8, 4, 10);
	else
		Gray_Display_DD(Gray.DD.Dark, 0xC0, 4, 0, 7);
}

void Gray_Pixel(int x, int y, Gray_Colour colour) {
	int pos;
	if (x & ~127 || y & ~63)
		return;
	pos = (y << 4) + (x >> 3);
	switch (colour) {
		case GRAY_BLACK:
			Gray.VRAM.Dark[pos] |= 128 >> (x & 7);
			Gray.VRAM.Light[pos] |= 128 >> (x & 7);
			break;
		case GRAY_WHITE:
			Gray.VRAM.Dark[pos] &= ~(128 >> (x & 7));
			Gray.VRAM.Light[pos] &= ~(128 >> (x & 7));
			break;
		case GRAY_XOR:
			Gray.VRAM.Dark[pos] ^= 128 >> (x & 7);
			Gray.VRAM.Light[pos] |= Gray.VRAM.Dark[pos] & (128 >> (x & 7));
			Gray.VRAM.Light[pos] &= Gray.VRAM.Dark[pos] | ~(128 >> (x & 7));
			break;
		case GRAY_CHECKER:
			if (y & 1 ^ x & 1) {
				Gray.VRAM.Dark[pos] &= ~(128 >> (x & 7));
				Gray.VRAM.Light[pos] &= ~(128 >> (x & 7));
			} else {
				Gray.VRAM.Dark[pos] |= 128 >> (x & 7);
				Gray.VRAM.Light[pos] |= 128 >> (x & 7);
			}
			break;
	}
}

void Gray_Vertical_Line(int x, int y1, int y2, Gray_Colour colour) {
	int i, j;
	char checker, byte;
	if (x & ~127 || (y1 < 0 && y2 < 0) || (y1 > 63 && y2 > 63))
		return;
	if (y1 > y2) {
		int tmp = y1;
		y1 = y2;
		y2 = tmp;
	}
	if (y1 < 0)
		y1 = 0;
	if (y2 > 63)
		y2 = 63;

	i = (y1 << 4) + (x >> 3);
	j = (y2 << 4) + (x >> 3);
	byte = 128 >> (x & 7);
	switch (colour) {
		case GRAY_WHITE:
			byte = ~byte;
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] &= byte;
				Gray.VRAM.Light[i] &= byte;
			}
			break;
		case GRAY_BLACK:
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] |= byte;
				Gray.VRAM.Light[i] |= byte;
			}
			break;
		case GRAY_XOR:
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] ^= byte;
				Gray.VRAM.Light[i] |= Gray.VRAM.Dark[i] & byte;
				Gray.VRAM.Light[i] &= Gray.VRAM.Dark[i] | ~byte;
			}
			break;
		case GRAY_CHECKER:
			checker = y1 & 1 ^ x & 1;
			for (; i <= j; i += 16) {
				if (checker) {
					Gray.VRAM.Dark[i] &= ~byte;
					Gray.VRAM.Light[i] &= ~byte;
				} else {
					Gray.VRAM.Dark[i] |= byte;
					Gray.VRAM.Light[i] |= byte;
				}
				checker = !checker;
			}
			break;
		case GRAY_LIGHT:
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] &= ~byte;
				Gray.VRAM.Light[i] |= byte;
			}
			break;
		case GRAY_DARK:
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] |= byte;
				Gray.VRAM.Light[i] &= ~byte;
			}
			break;
		case GRAY_INVERTED:
			for (; i <= j; i += 16) {
				Gray.VRAM.Dark[i] ^= byte;
				Gray.VRAM.Light[i] ^= byte;
			}
			break;
		case GRAY_CHECKER_GRAY:
			checker = y1 & 1 ^ x & 1;
			for (; i <= j; i += 16) {
				if (checker) {
					Gray.VRAM.Dark[i] &= ~byte;
					Gray.VRAM.Light[i] |= byte;
				} else {
					Gray.VRAM.Dark[i] |= byte;
					Gray.VRAM.Light[i] &= ~byte;
				}
				checker = !checker;
			}
			break;
	}
}

void Gray_Horizontal_Line(int y, int x1, int x2, Gray_Colour colour) {
	int i;
	char checker;
	if (y & ~63 || (x1 < 0 && x2 < 0) || (x1 > 127 && x2 > 127))
		return;
	if (x1 > x2) {
		i = x1;
		x1 = x2;
		x2 = i;
	}
	if (x1 < 0)
		x1 = 0;
	if (x2 > 127)
		x2 = 127;
	y <<= 4;
	switch (colour) {
		case GRAY_WHITE:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = 0;
					Gray.VRAM.Light[y + i] = 0;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
			}
			break;
		case GRAY_BLACK:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] |= 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] |= 255 >> (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] |= 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] |= 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = 255;
					Gray.VRAM.Light[y + i] = 255;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] |= (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] |= (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			}
			break;
		case GRAY_XOR:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] ^= 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] |= Gray.VRAM.Dark[y + (x1 >> 3)] & (255 >> (x1 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= Gray.VRAM.Dark[y + (x1 >> 3)] | ~(255 >> (x1 & 7));
				Gray.VRAM.Dark[y + (x2 >> 3)] ^= 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] |= Gray.VRAM.Dark[y + (x2 >> 3)] & (255 << 7 - (x2 & 7));
				Gray.VRAM.Light[y + (x2 >> 3)] &= Gray.VRAM.Dark[y + (x2 >> 3)] | ~(255 << 7 - (x2 & 7));
				for (i = (x1 >> 3) + 1; i < (x2 >> 3); i++) {
					Gray.VRAM.Dark[y + i] ^= 255;
					Gray.VRAM.Light[y + i] = Gray.VRAM.Dark[y + i];
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] ^= (255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] |= Gray.VRAM.Dark[y + (x1 >> 3)] & (255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= Gray.VRAM.Dark[y + (x1 >> 3)] | ~(255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
			}
			break;
		case GRAY_CHECKER:
			checker = (y & 16 ? 85 : 170);
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Dark[y + (x1 >> 3)] |= checker & 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] |= checker & 255 >> (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] |= checker & 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] |= checker & 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = checker;
					Gray.VRAM.Light[y + i] = checker;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Dark[y + (x1 >> 3)] |= checker & (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] |= checker & (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			}
			break;
		case GRAY_LIGHT:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] |= 255 >> (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] |= 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = 0;
					Gray.VRAM.Light[y + i] = 255;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] |= (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			}
			break;
		case GRAY_DARK:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] |= 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] |= 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = 255;
					Gray.VRAM.Light[y + i] = 0;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] |= (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
			}
			break;
		case GRAY_INVERTED:
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] ^= 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] ^= 255 >> (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] ^= 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] ^= 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < (x2 >> 3); i++) {
					Gray.VRAM.Dark[y + i] ^= 255;
					Gray.VRAM.Light[y + i] ^= 255;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] ^= (255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] ^= (255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
			}
			break;
		case GRAY_CHECKER_GRAY:
			checker = y & 16 ? 0x55 : 0xAA;
			if (x1 >> 3 != x2 >> 3) {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				Gray.VRAM.Dark[y + (x1 >> 3)] |= checker & 255 >> (x1 & 7);
				Gray.VRAM.Light[y + (x1 >> 3)] |= ~checker & 255 >> (x1 & 7);
				Gray.VRAM.Dark[y + (x2 >> 3)] |= checker & 255 << 7 - (x2 & 7);
				Gray.VRAM.Light[y + (x2 >> 3)] |= ~checker & 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++) {
					Gray.VRAM.Dark[y + i] = checker;
					Gray.VRAM.Light[y + i] = ~checker;
				}
			} else {
				Gray.VRAM.Dark[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				Gray.VRAM.Dark[y + (x1 >> 3)] |= checker & (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
				Gray.VRAM.Light[y + (x1 >> 3)] |= ~checker & (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			}
			break;
	}
}

void Gray_Rectangle(int x1, int y1, int x2, int y2, Gray_Colour colour) {
	Gray_Horizontal_Line(y1, x1, x2, colour);
	Gray_Horizontal_Line(y2, x1, x2, colour);
	Gray_Vertical_Line(x1 - 1, y1, y2, colour);
	Gray_Vertical_Line(x2 + 1, y1, y2, colour);
}

void Gray_BMP_Load(unsigned char* bmp, int x, int y, int width, int height, int fullWidthInBytes) {
	int i, j;
	unsigned char pixel;
	unsigned int pos;

	for (j = 0; j < height; j++) {
		for (i = 0; x + i < 128 && i < width; i++) {
			pos = (y + j << 4) + (x + i >> 3);

			pixel = 0x7F7F >> (x + i & 7);
			Gray.VRAM.Dark[pos] &= pixel;
			Gray.VRAM.Light[pos] &= pixel;

			pixel = bmp[(i >> 3) + (fullWidthInBytes * j)] & (0x80 >> (i & 7));
			pixel = (pixel >> (x & 7)) | (pixel << (8 - (x & 7)));
			Gray.VRAM.Dark[pos] |= pixel;
			Gray.VRAM.Light[pos] |= pixel;
		}
	}
}

void Gray_Mono_VRAM_To_Gray(void(*function)(void)) {
	char* vram = Gray_vram_adress();

	Gray_Copy_Buffer(Gray.VRAM.Dark, vram);
	(*function)();
	Gray_Copy_Buffer(vram, Gray.VRAM.Dark);

	Gray_Copy_Buffer(Gray.VRAM.Light, vram);
	(*function)();
	Gray_Copy_Buffer(vram, Gray.VRAM.Light);

}

//void Gray_Mono_VRAM_To_Gray(void(*function)(unsigned char* param1), unsigned char* param1) {
//	char* vram = Gray_vram_adress();
//
//	Gray_Copy_Buffer(Gray.VRAM.Dark, vram);
//	(*function)(param1);
//	Gray_Copy_Buffer(vram, Gray.VRAM.Dark);
//
//	Gray_Copy_Buffer(Gray.VRAM.Light, vram);
//	(*function)(param1);
//	Gray_Copy_Buffer(vram, Gray.VRAM.Light);
//
//}