#include "fxlib.h"
#include "stdio.h"
#include "timer.h"
#include "dispbios.h"
#include "Gray.c"
#include "Gray.h"
#include "MonochromeLib.c"
#include "MonochromeLib.h"
#include "KeyConvert.c"
#include "Syscall.h"

#define FIXEDPOINT 6
#define ZOOM_LIMIT 53
#define ABS(x) ((x)<0?(-(x)):(x))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))
#define CLAMP(x,min,max) ((x)<(min)?(min):(x)>(max)?(max):(x))

//typedef enum { FALSE, TRUE } Bool;

struct Mandelbrot {
	char Zoom;				//zoom level for graph
	Bool STOP;

	struct DoubleLength {
		int Upper;
		int Lower;
	} X, Y, Velocity;

} Mandel;

struct CameraView {
	unsigned char GrayTimings[7];
	unsigned char* FKeys;
	unsigned int Velocity;
	char Colour;				//Colour of camera rectangle: Black, White, Inverted, Transparent and Checkered: toggle with [F2]
	int Zoom;					//zoom
	int X, Y;
	Bool HUD;
	Bool GrayScale;
	Bool InstantUpdate;

} Camera = { 4,0,6,2,5,0,0 };

struct Iterations {
	int Selector;
	unsigned int Arguments[2];
	unsigned int Max;
	unsigned char* Buffer;
	short Gray[3];

} Iterations;

int abs64(int* upper, int* lower);

void neg64(int* upper, int* lower);

void sum64(int* upper, int* lower, int upperVelocity, int lowerVelocity);

void sub64(int* upper, int* lower, int upperVelocity, int lowerVelocity);

void drawMandel(int zoom, int* iterationsBuffer, int upperY, int lowerY, int upperX, int lowerX, int maxIterations, int* STOP);

int pow(int number, int power) {
	int i;
	int output = 1;
	for (i = 0; i < power; i++)
		output *= number;
	return output;
}

void shift64(unsigned int* high, unsigned int* low, int shift) {
	if (shift >= 64 || shift <= -64) {
		*high = 0;
		*low = 0;
	} else if (shift >= 32) {
		*high = *low << shift - 32;
		*low = 0;
	} else if (shift <= -32) {
		*low = *high << shift + 32;
		*high = 0;
	} else if (shift > 0) {
		*high = (*high << shift) | (*low >> 32 - shift);
		*low = *low << shift;
	} else if (shift < 0) {
		*low = (*low << shift) | (*high << 32 + shift);
		*high = *high << shift;
	}
}

void sprintFP(unsigned char* string, unsigned int fixedPoint, unsigned int high, unsigned int low) {
	unsigned char fraction[32];
	int tempHigh = high;
	int i = 0;

	abs64(&high, &low);
	shift64(&high, &low, fixedPoint - 4);
	low >>= 4;

	high &= 0x0FFFFFFF;
	low &= 0x0FFFFFFF;
	do {
		high *= 10;
		low *= 10;
		high += low >> 28;
		fraction[i++] = '0' + (high >> 28);
		high &= 0x0FFFFFFF;
		low &= 0x0FFFFFFF;
	} while ((high || low) && i < 27);
	fraction[i] = '\0';

	sprintf(string, "%s%s%u.%s", string, tempHigh < 0 ? "-" : "+", ABS(tempHigh) >> 32 - fixedPoint, fraction);
}

void stop(void) {					//stops drawing set if user presses [EXIT] or [MENU]
	int kcode1 = 0, kcode2 = 0;		//row & col keycode for Bkey_GetKeyWait()
	short unused;					//unused

	Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTOFF_TIMEROFF, 0, !FALSE, &unused);
	if ((kcode1 == 4 && (kcode2 == 8 || kcode2 == 9)) || (kcode1 == 1 && kcode2 == 1)) {
		//	KEY_CTRL_MENU || KEY_CTRL_EXIT || KEY_CTRL_AC
		Mandel.STOP = TRUE;
		KillTimer(1);
	}
}

void drawGrayTuneText() {
	unsigned char string[32];
	int i;
	for (i = 1; i < 7; i++) {
		sprintf(&string, "%u", Camera.GrayTimings[i]);
		PrintMini(-4 + i * 5, 1, string, Camera.GrayTimings[0] == i ? MINI_REV : MINI_OVER);
		if (Camera.GrayTimings[0] != i)
			ML_vertical_line(-5 + i * 5, 0, 6, ML_WHITE);
	}
}

void drawIterationTuneText() {
	unsigned char string[32];
	sprintf(&string, "LIGHT:%d", Iterations.Gray[1]);
	PrintMini(1, 1, string, Iterations.Gray[0] == 1 ? MINI_REV : MINI_OVER);
	sprintf(&string, "DARK:%d", Iterations.Gray[2]);
	PrintMini(4, 8, string, Iterations.Gray[0] == 2 ? MINI_REV : MINI_OVER);
	ML_vertical_line(0, 0, 6, Iterations.Gray[0] == 1 ? ML_BLACK : ML_WHITE);
	ML_vertical_line(3, 7, 13, Iterations.Gray[0] == 2 ? ML_BLACK : ML_WHITE);
}

void drawGrayMandelbrot() {
	unsigned char byteDark = 0, byteLight = 0, byte;
	char *tempBufferDark = Gray.VRAM.Dark, *tempBufferLight = Gray.VRAM.Light;
	int x, y;

	for (y = 0; y < 64; y++) {
		for (x = 0; x < 128; x++) {
			byte = Iterations.Buffer[(y << 7) + x];
			byteDark = (byteDark << 1) | (byte <= Iterations.Gray[2]);
			byteLight = (byteLight << 1) | ((byte <= Iterations.Gray[1]) != (byte <= Iterations.Gray[2])) | (byte == 0);
			if ((x & 7) == 7) {
				*tempBufferDark++ = byteDark;
				*tempBufferLight++ = byteLight;
			}
		}
	}
}

void updateVelocity() {
	Camera.Velocity = Camera.Zoom > 4 ? 1 : 16 >> Camera.Zoom;

	Mandel.Velocity.Upper = 0x80000000;
	Mandel.Velocity.Lower = 0x00000000;
	shift64(&Mandel.Velocity.Upper, &Mandel.Velocity.Lower, -(Camera.Zoom > 4 ? Mandel.Zoom - Camera.Zoom + 4 : Mandel.Zoom) - FIXEDPOINT);
	if (!Mandel.Velocity.Upper && !Mandel.Velocity.Lower)
		Mandel.Velocity.Lower = 0x00000001;
}

void drawHUD() {
	unsigned char string[32];
	int x, y, i;

	neg64(&Mandel.X.Upper, &Mandel.X.Lower);
	sprintf(&string, "X:", 0);
	sprintFP(&string, FIXEDPOINT, Mandel.X.Upper, Mandel.X.Lower);
	PrintMini(0, 0, string, 0);
	neg64(&Mandel.X.Upper, &Mandel.X.Lower);

	sprintf(&string, "Y:", 0);
	sprintFP(&string, FIXEDPOINT, Mandel.Y.Upper, Mandel.Y.Lower);
	PrintMini(0, 6, string, 0);

	sprintf(&string, "MaxI:%u", Iterations.Arguments[0] * Mandel.Zoom + Iterations.Arguments[1]);
	PrintMini(0, 44, string, 0);

	if (Mandel.Zoom > 32)
		sprintf(&string, "Zoom:2^%ux", Mandel.Zoom - 1);
	else if (Mandel.Zoom > 0)
		sprintf(&string, "Zoom:%ux", 1 << Mandel.Zoom - 1);
	else
		sprintf(&string, "Zoom:0.5x", 0);

	PrintMini(0, 50, string, 0);
}

void drawCameraRectangle() {
	if (Camera.Zoom > -26) {
		if (Camera.GrayScale == TRUE) {
			if (Camera.Zoom < 8)
				Gray_Rectangle(65 - (128 >> Camera.Zoom) + Camera.X, 32 - (Camera.Zoom > 6 ? 1 : (64 >> Camera.Zoom)) + Camera.Y, 62 + (128 >> Camera.Zoom) + Camera.X, 31 + (Camera.Zoom > 6 ? 0 : (64 >> Camera.Zoom)) + Camera.Y, Camera.Colour);
			else
				Gray_Pixel(Camera.X + 64, Camera.Y + 31, Camera.Colour);
		} else {
			if (Camera.Zoom < 8)
				ML_rectangle(65 - (128 >> Camera.Zoom) + Camera.X, 32 - (Camera.Zoom > 6 ? 1 : (64 >> Camera.Zoom)) + Camera.Y, 62 + (128 >> Camera.Zoom) + Camera.X, 31 + (Camera.Zoom > 6 ? 0 : (64 >> Camera.Zoom)) + Camera.Y, Camera.Colour);
			else
				ML_pixel(Camera.X + 64, Camera.Y + 31, Camera.Colour);
		}
	}
}

void updateScreen() {
	unsigned char string[32];
	static shouldUpdateHUD = -1;
	int i;

	RestoreDisp(SAVEDISP_PAGE1);

	if (Camera.GrayScale == shouldUpdateHUD) {
		drawCameraRectangle();
		if (Camera.HUD == TRUE) {
			for (i = 0; i < 6; i++) {
				ML_BMP_Load(Camera.FKeys + i * 24, i * 21, 56, 23, 8, 3);
			}
			if (Camera.InstantUpdate == TRUE)
				ML_BMP_Load(Camera.FKeys + 6 * 24, 5 * 21, 56, 23, 8, 3);
			drawHUD();
		}
	}

	if (Camera.GrayScale == TRUE) {
		drawGrayMandelbrot();
		if (shouldUpdateHUD == TRUE) {
			drawCameraRectangle();
			if (Camera.HUD == TRUE) {
				Gray_Mono_VRAM_To_Gray(drawHUD);

				for (i = 0; i < 6; i++) {
					Gray_BMP_Load(Camera.FKeys + i * 24, i * 21, 56, 23, 8, 3);
				}
				if (Camera.InstantUpdate == TRUE)
					Gray_BMP_Load(Camera.FKeys + 6 * 24, 5 * 21, 56, 23, 8, 3);
			}
		}
		Gray_Swap_Buffers();

		shouldUpdateHUD = TRUE;
	} else {
		ML_display_vram();

		shouldUpdateHUD = FALSE;
	}
}

void drawMandelbrot() {
	char* vram = ML_vram_adress();
	int kcode1 = 0, kcode2 = 0;				//row & col keycode for Bkey_GetKeyWait()
	unsigned int i;
	unsigned char byte = 0;
	unsigned int upperX = Mandel.X.Upper, lowerX = Mandel.X.Lower;
	unsigned int upperY = Mandel.Y.Upper, lowerY = Mandel.Y.Lower;
	int offsetHigh = 0x80000000;
	int offsetLow = 0x00000000;

	Gray_Stop();

	shift64(&offsetHigh, &offsetLow, -Mandel.Zoom - FIXEDPOINT + 3);
	sum64(&upperX, &lowerX, offsetHigh, offsetLow);

	shift64(&offsetHigh, &offsetLow, -1);
	sum64(&upperY, &lowerY, offsetHigh, offsetLow);

	if (Iterations.Max = Iterations.Arguments[0] * Mandel.Zoom + Iterations.Arguments[1]) {
		Mandel.STOP = FALSE;
		RestoreDisp(SAVEDISP_PAGE1);
		ML_display_vram();
		memset(Iterations.Buffer, 0xFF, 8192);
		neg64(&upperX, &lowerX);
		neg64(&upperY, &lowerY);
		SetTimer(ID_USER_TIMER1, 200, stop);
		drawMandel(Mandel.Zoom, Iterations.Buffer, upperY, lowerY, upperX, lowerX, Iterations.Max, &Mandel.STOP);
		KillTimer(ID_USER_TIMER1);
		for (i = 0; i < 8192; i++) {
			byte <<= 1;
			byte |= Iterations.Buffer[i] == 0;
			if ((i & 7) == 7)
				*vram++ = byte;
		}
		Keyboard_ClrBuffer();
	} else
		ML_clear_vram();

	SaveDisp(SAVEDISP_PAGE1);
	if (Camera.GrayScale == TRUE) {
		drawGrayMandelbrot();
		Gray_Swap_Buffers();
		Gray_Start();
	} else
		ML_display_vram();
	Camera.X = 0;
	Camera.Y = 0;
	Camera.Zoom = 1;
	updateVelocity();
}

void reset() {
	KillTimer(ID_USER_TIMER1);
	Iterations.Selector = 0;
	Iterations.Arguments[0] = 8;
	Iterations.Arguments[1] = 24;
	Iterations.Gray[0] = 1;
	Iterations.Gray[1] = 30;
	Iterations.Gray[2] = 24;
	Camera.InstantUpdate = FALSE;
	Camera.HUD = TRUE;
	Camera.GrayScale = FALSE;
	Camera.Colour = GRAY_XOR;
	Camera.X = Camera.Y = 0;
	Camera.Zoom = 1;
	Mandel.Zoom = 1;
	Mandel.X.Upper = Mandel.X.Lower = 0;
	Mandel.Y.Upper = Mandel.Y.Lower = 0;
	Mandel.STOP = FALSE;
	updateVelocity();
	updateScreen();
	ML_clear_vram();
	ML_display_vram();
	SaveDisp(SAVEDISP_PAGE1);
	if (Gray.Initialised == TRUE)
		Gray_Init();
	drawMandelbrot();
}

int systemKeys(int key) {
	int kcode1 = 0, kcode2 = 0;					//row & col keycode for Bkey_GetKeyWait()
	short matrixCode;
	short unused;								//unused
	Bool repeat;

	do {
		repeat = FALSE;
		switch (key) {
			case KEY_CTRL_MENU:
				if (Camera.GrayScale == TRUE) {
					Camera.GrayScale = Gray_Stop() == 0 ? 2 : FALSE;
					matrixCode = 0x308;
					Keyboard_PutMatrixCode(&matrixCode);
				}
				break;
			case KEY_CTRL_EXIT:
				if (!Gray_Stop() == TRUE) {
					Camera.GrayScale = FALSE;
					updateScreen();
				} else {
					SaveDisp(SAVEDISP_PAGE2);
					ML_clear_vram();
					PrintXY(6, 28, "Press [EXE] to Quit", 0);
					do GetKey(&key);
					while (key == KEY_CTRL_SHIFT || key == KEY_CTRL_EXIT);
					key = key == KEY_CTRL_EXE ? KEY_CTRL_QUIT : KEY_CTRL_EXE;
					RestoreDisp(SAVEDISP_PAGE2);
					ML_display_vram();
				}
				break;
			case KEY_CTRL_AC:
				reset();
				break;
			case KEY_CTRL_SHIFT:
				Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTON_TIMEROFF, 0, !FALSE, &unused);
				key = KeyEncode(kcode1, kcode2);
				switch (key) {
					case KEY_CTRL_AC:
						Gray_PowerOffCalc(TRUE);
						if (Camera.GrayScale == FALSE)
							ML_display_vram();
						break;
					case KEY_CTRL_LEFT:
						Keyboard_ClrBuffer();
						matrixCode = 0x608;			//KEY_CTRL_SHIFT
						Keyboard_PutMatrixCode(&matrixCode);
						GetKey(&key);
						matrixCode = 0x208;			//KEY_CTRL_LEFT
						Keyboard_PutMatrixCode(&matrixCode);
						GetKey(&key);
						if (key != KEY_CTRL_SHIFT) {
							repeat = TRUE;
						}
						break;
					case KEY_CTRL_RIGHT:
						Keyboard_ClrBuffer();
						matrixCode = 0x608;			//KEY_CTRL_SHIFT
						Keyboard_PutMatrixCode(&matrixCode);
						GetKey(&key);
						matrixCode = 0x107;			//KEY_CTRL_RIGHT
						Keyboard_PutMatrixCode(&matrixCode);
						GetKey(&key);
						if (key != KEY_CTRL_SHIFT) {
							repeat = TRUE;
						}
						break;
					case KEY_CTRL_EXIT:
						Gray_Stop();
						RestoreDisp(SAVEDISP_PAGE1);
						key = KEY_CTRL_QUIT;
						break;
				}
				break;
			case KEY_CTRL_ALPHA:
				Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTON_TIMEROFF, 0, !FALSE, &unused);
				key = KeyEncode(kcode1, kcode2);
				switch (key) {
					case KEY_CTRL_AC:
						ML_clear_display();
						Gray_PowerOffCalc(FALSE);
						break;
					case KEY_CTRL_EXIT:
						Gray_Stop();
						ML_clear_display();
						key = KEY_CTRL_QUIT;
				}
				break;
			default:
				key = KEY_CTRL_NOP;
				break;
		}
	} while (repeat == TRUE);
	return key;
}

int keyInput() {
	char* tempBufferDark = Gray.VRAM.Dark;
	char* tempBufferLight = Gray.VRAM.Light;
	int kcode1 = 0, kcode2 = 0;					//row & col keycode for Bkey_GetKeyWait()
	short unused;								//unused
	unsigned int key = 0;						//Get key being pressed
	unsigned char string[32];					//Used in converting int to string
	int x, y, i;
	unsigned char temp[10];

	Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTON_TIMEROFF, 0, !TRUE, &unused);
	key = KeyEncode(kcode1, kcode2);

	if (Camera.GrayScale == 2)
		Camera.GrayScale = Gray_Start();

	switch (key) {
		case KEY_CHAR_PLUS:
			if (Mandel.Zoom <= ZOOM_LIMIT) {
				Mandel.Zoom++;
				Camera.Zoom++;
				updateVelocity();
				if (Camera.InstantUpdate == TRUE)
					drawMandelbrot();
				else
					updateScreen();
			}
			break;
		case KEY_CHAR_MINUS:
			if (Mandel.Zoom) {
				Mandel.Zoom--;
				Camera.Zoom--;
				updateVelocity();
				if (Camera.InstantUpdate == TRUE)
					drawMandelbrot();
				else
					updateScreen();
			}
			break;
		case KEY_CTRL_UP:
			Camera.Y -= Camera.Velocity;
			sum64(&Mandel.Y.Upper, &Mandel.Y.Lower, Mandel.Velocity.Upper, Mandel.Velocity.Lower);
			if (Camera.InstantUpdate == TRUE)
				drawMandelbrot();
			else
				updateScreen();
			break;
		case KEY_CTRL_DOWN:
			Camera.Y += Camera.Velocity;
			sub64(&Mandel.Y.Upper, &Mandel.Y.Lower, Mandel.Velocity.Upper, Mandel.Velocity.Lower);
			if (Camera.InstantUpdate == TRUE)
				drawMandelbrot();
			else
				updateScreen();
			break;
		case KEY_CTRL_LEFT:
			Camera.X -= Camera.Velocity;
			sum64(&Mandel.X.Upper, &Mandel.X.Lower, Mandel.Velocity.Upper, Mandel.Velocity.Lower);
			if (Camera.InstantUpdate == TRUE)
				drawMandelbrot();
			else
				updateScreen();
			break;
		case KEY_CTRL_RIGHT:
			Camera.X += Camera.Velocity;
			sub64(&Mandel.X.Upper, &Mandel.X.Lower, Mandel.Velocity.Upper, Mandel.Velocity.Lower);
			if (Camera.InstantUpdate == TRUE)
				drawMandelbrot();
			else
				updateScreen();
			break;
		case KEY_CTRL_F1:
			Camera.HUD = !Camera.HUD;
			updateScreen();
			break;
		case KEY_CTRL_F2:
			if (Camera.Colour < GRAY_CHECKER)
				Camera.Colour++;
			else
				Camera.Colour = GRAY_TRANSPARENT;
			updateScreen();
			break;
		case KEY_CTRL_F3:
			Gray_Stop();
			SaveDisp(SAVEDISP_PAGE2);
			do {
				PrintXY(6, 14, "MaxI = A * Zoom + B", 0);
				sprintf(&string, "A:%u", Iterations.Arguments[0]);
				PrintXY(48, 24, &string, 0);
				sprintf(&string, "B:%u", Iterations.Arguments[1]);
				PrintXY(48, 32, &string, 0);
				sprintf(&string, "MaxI:%u", Iterations.Arguments[0] * Mandel.Zoom + Iterations.Arguments[1]);
				PrintXY(30, 42, &string, 0);

				GetKey(&key);
				switch (key) {
					case KEY_CTRL_UP:
						Iterations.Selector = 0;
						break;
					case KEY_CTRL_DOWN:
						Iterations.Selector = 1;
						break;
					case KEY_CHAR_MINUS:
						if (Iterations.Arguments[Iterations.Selector])
							Iterations.Arguments[Iterations.Selector]--;
						break;
					case KEY_CHAR_PLUS:
						Iterations.Arguments[Iterations.Selector]++;
						break;
					case KEY_CTRL_DEL:
						Iterations.Arguments[Iterations.Selector] /= 10;
						break;
					case KEY_CTRL_AC:
						Iterations.Arguments[Iterations.Selector] = 0;
						break;
					case KEY_CTRL_ALPHA:
						GetKey(&key);
						switch (key) {
							case KEY_CTRL_AC:
								ML_clear_display();
								Gray_PowerOffCalc(FALSE);
								break;
							case KEY_CTRL_EXIT:
								Gray_Stop();
								ML_clear_display();
								key = KEY_CTRL_QUIT;
						}
						break;
					default:
						if (key >= '0' && key <= '9') {
							Iterations.Arguments[Iterations.Selector] = Iterations.Arguments[Iterations.Selector] * 10 + key - '0';
						}
						break;
				}

				Iterations.Arguments[Iterations.Selector] = MIN(Iterations.Arguments[Iterations.Selector], 32768);

				RestoreDisp(SAVEDISP_PAGE2);
			} while (key != KEY_CTRL_EXIT && key != KEY_CTRL_EXE && key != KEY_CTRL_F3 && key != KEY_CTRL_QUIT);
			if (Camera.GrayScale == TRUE)
				Gray_Start();
			else
				ML_display_vram();
			break;
		case KEY_CTRL_F4:
			Camera.GrayScale = Gray_Toggle();
			if (Camera.GrayScale) {
				do {
					tempBufferDark = Gray.VRAM.Dark;
					tempBufferLight = Gray.VRAM.Light;
					for (y = 0; y < 64; y++) {
						for (x = 0; x < 16; x++) {
							if (x < 8)
								*tempBufferDark++ = 0xFF;
							else
								*tempBufferDark++ = 0;
							if (x < 4 || (x > 7 && x < 12))
								*tempBufferLight++ = 0xFF;
							else
								*tempBufferLight++ = 0;
						}
					}
					Gray_Mono_VRAM_To_Gray(drawGrayTuneText);
					Gray_Swap_Buffers();
					Gray.Time = 0;
					for (i = 1; i < 7; i++)
						Gray.Time = Gray.Time * 10 + Camera.GrayTimings[i];
					Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTON_TIMEROFF, 0, !TRUE, &unused);
					key = KeyEncode(kcode1, kcode2);
					if (Camera.GrayScale == 2)
						Camera.GrayScale = Gray_Start();
					switch (key) {
						case KEY_CTRL_UP:
							if (Camera.GrayTimings[Camera.GrayTimings[0]] == 9)
								Camera.GrayTimings[Camera.GrayTimings[0]] = 0;
							else
								Camera.GrayTimings[Camera.GrayTimings[0]]++;
							break;
						case KEY_CHAR_PLUS:
							Gray.Time += pow(10, 6 - Camera.GrayTimings[0]);
							Gray.Time = MIN(Gray.Time, 999999);
							sprintf(Camera.GrayTimings, "%u%06u", Camera.GrayTimings[0], Gray.Time);
							for (i = 0; i < 7; i++)
								Camera.GrayTimings[i] -= 0x30;
							break;
						case KEY_CTRL_DOWN:
							if (Camera.GrayTimings[Camera.GrayTimings[0]] == 0)
								Camera.GrayTimings[Camera.GrayTimings[0]] = 9;
							else
								Camera.GrayTimings[Camera.GrayTimings[0]]--;
							break;
						case KEY_CHAR_MINUS:
							Gray.Time -= pow(10, 6 - Camera.GrayTimings[0]);
							Gray.Time = MAX(Gray.Time, 0);
							sprintf(Camera.GrayTimings, "%u%06u", Camera.GrayTimings[0], Gray.Time);
							for (i = 0; i < 7; i++)
								Camera.GrayTimings[i] -= 0x30;
							break;
						case KEY_CTRL_LEFT:
						case KEY_CHAR_MULT:
							Camera.GrayTimings[0] = MAX(Camera.GrayTimings[0]--, 1);
							break;
						case KEY_CTRL_RIGHT:
						case KEY_CHAR_DIV:
							Camera.GrayTimings[0] = MIN(Camera.GrayTimings[0]++, 6);
							break;
						case KEY_CTRL_DEL:
							key = KEY_CHAR_0;
						case KEY_CHAR_0:
						case KEY_CHAR_1:
						case KEY_CHAR_2:
						case KEY_CHAR_3:
						case KEY_CHAR_4:
						case KEY_CHAR_5:
						case KEY_CHAR_6:
						case KEY_CHAR_7:
						case KEY_CHAR_8:
						case KEY_CHAR_9:
							Camera.GrayTimings[Camera.GrayTimings[0]] = key - '0';
							break;
						case KEY_CTRL_F4:
						case KEY_CTRL_EXE:
						case KEY_CTRL_EXIT:
							break;
						case KEY_CTRL_AC:
							for (i = 1; i < 7; i++)
								Camera.GrayTimings[i] = 0;
							break;
						default:
							key = systemKeys(key);
							break;
					}
				} while (key != KEY_CTRL_F4 && key != KEY_CTRL_EXE && key != KEY_CTRL_EXIT && key != KEY_CTRL_QUIT);
				while (key != KEY_CTRL_EXIT && key != KEY_CTRL_QUIT) {
					Iterations.Gray[1] = CLAMP(Iterations.Gray[1], 0, 255);
					Iterations.Gray[2] = CLAMP(Iterations.Gray[2], 0, 255);
					drawGrayMandelbrot();
					Gray_Mono_VRAM_To_Gray(drawIterationTuneText);
					Gray_Swap_Buffers();
					Bkey_GetKeyWait(&kcode1, &kcode2, KEYWAIT_HALTON_TIMEROFF, 0, !TRUE, &unused);
					key = KeyEncode(kcode1, kcode2);
					if (Camera.GrayScale == 2)
						Camera.GrayScale = Gray_Start();
					switch (key) {
						case KEY_CTRL_UP:
						case KEY_CHAR_MULT:
							Iterations.Gray[0] = 1;
							break;
						case KEY_CTRL_DOWN:
						case KEY_CHAR_DIV:
							Iterations.Gray[0] = 2;
							break;
						case KEY_CHAR_PLUS:
						case KEY_CTRL_RIGHT:
							Iterations.Gray[Iterations.Gray[0]] += 1;
							break;
						case KEY_CHAR_MINUS:
						case KEY_CTRL_LEFT:
							Iterations.Gray[Iterations.Gray[0]] -= 1;
							break;
						case KEY_CTRL_DEL:
							Iterations.Gray[Iterations.Gray[0]] /= 10;
							break;
						case KEY_CHAR_0:
						case KEY_CHAR_1:
						case KEY_CHAR_2:
						case KEY_CHAR_3:
						case KEY_CHAR_4:
						case KEY_CHAR_5:
						case KEY_CHAR_6:
						case KEY_CHAR_7:
						case KEY_CHAR_8:
						case KEY_CHAR_9:
							Iterations.Gray[Iterations.Gray[0]] = Iterations.Gray[Iterations.Gray[0]] * 10 + key - '0';
							break;
						case KEY_CTRL_F4:
						case KEY_CTRL_EXE:
							key = KEY_CTRL_EXIT;
						case KEY_CTRL_EXIT:
							break;
						case KEY_CTRL_AC:
							Iterations.Gray[1] = 0;
							Iterations.Gray[2] = 0;
							break;
						default:
							key = systemKeys(key);
							break;
					}
				}
			}
			updateScreen();
			break;
		case KEY_CTRL_F5:
			updateScreen();
			neg64(&Mandel.X.Upper, &Mandel.X.Lower);
			sprintf(&string, "X:0x%08X%08X", Mandel.X.Upper, Mandel.X.Lower);
			PrintMini(0, 20, string, 0);
			neg64(&Mandel.X.Upper, &Mandel.X.Lower);
			sprintf(&string, "Y:0x%08X%08X", Mandel.Y.Upper, Mandel.Y.Lower);
			PrintMini(0, 28, string, 0);
			ML_display_vram();
			break;
		case KEY_CTRL_F6:
			Camera.InstantUpdate = !Camera.InstantUpdate;
			updateScreen();
			break;
		case KEY_CTRL_EXE:
			drawMandelbrot();
			break;
		case KEY_CTRL_VARS:
			ML_display_vram();
			break;
		default:
			key = systemKeys(key);
			if (!key)
				updateScreen();
			break;
	}

	return key;
}

int AddIn_main(int isAppli, unsigned short OptionNum) {		//Main function
	/*
	64bit Fixed Point number format - spilt between two 32bit variables
	1:5:58
	Sign:int:Frac
	:00000:0000000000000000000000000000000000000000000000000000000000
	00000.00000000000000000000000000,00000000000000000000000000000000
	high = 00000.00000000000000000000000000
	low  = 00000000000000000000000000000000
	*/

	/*
	unsigned char string[32];
	sprintf(&string, "%d", );
	PrintXY(30, 30, string, 0);
	*/

	//char* vram = ML_vram_adress();
	int i;
	unsigned int key;
	unsigned char string[32];
	unsigned char iterations_Buffer[8192];
	unsigned char FKeys[168] = {
		0x3F,0xFF,0xF8, 0x20,0x00,0x00, 0x24,0xA5,0xC0, 0x24,0xA5,0x20, 0x27,0xA5,0x20, 0x24,0xA5,0x20, 0x24,0x99,0xC0, 0x20,0x00,0x00,	//HUD		Heads Up Display
		0x3F,0xFF,0xF8, 0x20,0x00,0x00, 0x24,0x91,0x30, 0x29,0x52,0xA8, 0x29,0x52,0xB0, 0x29,0x52,0xA8, 0x24,0x99,0x28, 0x20,0x00,0x00,	//COLOR		Colour
		0x3F,0xFF,0xF8, 0x2E,0xDA,0x88, 0x24,0xAA,0xD8, 0x2A,0xAD,0xD8, 0x2A,0x8A,0xD8, 0x2E,0xAA,0xD0, 0x2E,0xAA,0x80,	0x3F,0xFF,0xC0,	//MAXI		Max Iterations
		0x3F,0xFF,0xF8, 0x31,0x3B,0x58, 0x2F,0x55,0x58, 0x29,0x55,0x18, 0x2D,0x31,0xB8, 0x2D,0x55,0xB0, 0x31,0x55,0xA0, 0x3F,0xFF,0xC0,	//GRAY		Gray Scale
		0x3F,0xFF,0xF8, 0x37,0xEE,0x38, 0x37,0xED,0xF8, 0x37,0xDE,0x78, 0x37,0xDF,0xB8, 0x37,0xBF,0xB0, 0x31,0xBC,0x60, 0x3F,0xFF,0xC0,	//L/S		Load/Save
		0x3F,0xFF,0xF8, 0x20,0x00,0x00, 0x24,0x49,0x20, 0x26,0xD5,0xA0, 0x25,0x5D,0x60, 0x24,0x55,0x20, 0x24,0x55,0x20, 0x20,0x00,0x00,	//MAN		Manual
		0x3F,0xFF,0xF8, 0x3C,0x70,0x78, 0x3D,0xBD,0xF8, 0x3D,0xBD,0xF8, 0x3C,0x7D,0xF8, 0x3D,0xBD,0xF8, 0x3D,0xBD,0xF8, 0x3F,0xFF,0xF8,	//RT		Real Time
		//0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8,	//Full Black
		//0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF8, 0x3F,0xFF,0xF0, 0x3F,0xFF,0xE0, 0x3F,0xFF,0xC0,	//Black with missing corner
		//0x3F,0xFF,0xF8, 0x20,0x00,0x00, 0x20,0x00,0x00, 0x20,0x00,0x00, 0x20,0x00,0x00, 0x20,0x00,0x00, 0x20,0x00,0x00, 0x20,0x00,0x00,	//White with Black border
	};

	Iterations.Buffer = iterations_Buffer;
	Camera.FKeys = FKeys;

	reset();

	do {
		key = keyInput();

	} while (key != KEY_CTRL_QUIT);

	RestoreDisp(SAVEDISP_PAGE1);
	return 1;
}


#pragma section _BR_Size
unsigned long BR_Size;
#pragma section
#pragma section _TOP
int InitializeSystem(int isAppli, unsigned short OptionNum) {
	return INIT_ADDIN_APPLICATION(isAppli, OptionNum);
}
#pragma section