#include "MonochromeLib.h"
#include "Syscall.h"

void ML_set_vram() {
	char* vram = ML_vram_adress();
	int i;
	for (i = 0; i < 1024; i++)
		*vram++ = 0xFF;
}

void ML_clear_vram() {
	int i, end, * pointer_long, vram;
	char* pointer_byte;
	vram = (int)ML_vram_adress();
	end = 4 - vram & 3;
	pointer_byte = (char*)vram;
	for (i = 0; i < end; i++)
		pointer_byte[i] = 0;
	pointer_long = (int*)(vram + end);
	for (i = 0; i < 255; i++)
		pointer_long[i] = 0;
	pointer_byte += 1020 + end;
	end = vram & 3;
	for (i = 0; i < end; i++)
		pointer_byte[i] = 0;
}

void ML_display_vram() {
	char* LCD_register_selector = (char*)0xB4000000;
	char* LCD_data_register = (char*)0xB4010000;
	char* vram = ML_vram_adress();
	unsigned char OS[10];
	int i, j;
	
	System_GetOSVersion(OS);

	if (OS[1] == '3') {
		for (i = 0; i < 64; i++) {
			*LCD_register_selector = 8;
			*LCD_data_register = i | 128;
			*LCD_register_selector = 8;
			*LCD_data_register = 4;
			*LCD_register_selector = 10;
			for (j = 0; j < 16; j++)
				*LCD_data_register = *vram++;
		}
	} else {
		for (i = 0; i < 64; i++) {
			*LCD_register_selector = 4;
			*LCD_data_register = i | 192;
			*LCD_register_selector = 4;
			*LCD_data_register = 0;
			*LCD_register_selector = 7;
			for (j = 0; j < 16; j++)
				*LCD_data_register = *vram++;
		}
	}
}

void ML_clear_display() {
	char* LCD_register_selector = (char*)0xB4000000, * LCD_data_register = (char*)0xB4010000, * vram;
	int i, j;

	for (i = 0; i < 64; i++) {
		*LCD_register_selector = 4;
		*LCD_data_register = i | 192;
		*LCD_register_selector = 4;
		*LCD_data_register = 0;
		*LCD_register_selector = 7;
		for (j = 0; j < 16; j++)
			*LCD_data_register = 0;
	}
}

void ML_BMP_Load(unsigned char* bmp, int x, int y, int width, int height, int fullWidthInBytes) {
	char* vram = ML_vram_adress();
	unsigned char pixel;
	int i, j;
	int pos;

	for (j = 0; j < height; j++) {
		for (i = 0; x + i < 128 && i < width; i++) {
			pos = (y + j << 4) + (x + i >> 3);

			pixel = 0x7F7F >> (x + i & 7);
			vram[pos] &= pixel;

			pixel = bmp[(i >> 3) + (fullWidthInBytes * j)] & (0x80 >> (i & 7));
			pixel = (pixel >> (x & 7)) | (pixel << (8 - (x & 7)));
			vram[pos] |= pixel;
		}
	}
}

void ML_pixel(int x, int y, ML_Colour colour) {
	char* vram = ML_vram_adress();
	if (x & ~127 || y & ~63) return;
	switch (colour) {
		case ML_BLACK:
			vram[(y << 4) + (x >> 3)] |= 128 >> (x & 7);
			break;
		case ML_WHITE:
			vram[(y << 4) + (x >> 3)] &= ~(128 >> (x & 7));
			break;
		case ML_XOR:
			vram[(y << 4) + (x >> 3)] ^= 128 >> (x & 7);
			break;
		case ML_CHECKER:
			if (y & 1 ^ x & 1) vram[(y << 4) + (x >> 3)] &= ~(128 >> (x & 7));
			else vram[(y << 4) + (x >> 3)] |= 128 >> (x & 7);
			break;
	}
}

void ML_vertical_line(int x, int y1, int y2, ML_Colour colour) {
	int i, j;
	char checker, byte, * vram = ML_vram_adress();
	if (x & ~127 || (y1 < 0 && y2 < 0) || (y1 > 63 && y2 > 63)) return;
	if (y1 > y2) {
		int tmp = y1;
		y1 = y2;
		y2 = tmp;
	}
	if (y1 < 0) y1 = 0;
	if (y2 > 63) y2 = 63;

	i = (y1 << 4) + (x >> 3);
	j = (y2 << 4) + (x >> 3);
	switch (colour) {
		case ML_BLACK:
			byte = 128 >> (x & 7);
			for (; i <= j; i += 16)
				vram[i] |= byte;
			break;
		case ML_WHITE:
			byte = ~(128 >> (x & 7));
			for (; i <= j; i += 16)
				vram[i] &= byte;
			break;
		case ML_XOR:
			byte = 128 >> (x & 7);
			for (; i <= j; i += 16)
				vram[i] ^= byte;
			break;
		case ML_CHECKER:
			byte = 128 >> (x & 7);
			checker = y1 & 1 ^ x & 1;
			for (; i <= j; i += 16) {
				if (checker) vram[i] &= ~byte;
				else vram[i] |= byte;
				checker = !checker;
			}
			break;
	}
}

void ML_horizontal_line(int y, int x1, int x2, ML_Colour colour) {
	int i;
	char checker;
	char* vram = ML_vram_adress();
	if (y & ~63 || (x1 < 0 && x2 < 0) || (x1 > 127 && x2 > 127))
		return;
	if (x1 > x2) {
		i = x1;
		x1 = x2;
		x2 = i;
	}
	if (x1 < 0)
		x1 = 0;
	if (x2 > 127)
		x2 = 127;
	switch (colour) {
		case ML_BLACK:
			if (x1 >> 3 != x2 >> 3) {
				vram[(y << 4) + (x1 >> 3)] |= 255 >> (x1 & 7);
				vram[(y << 4) + (x2 >> 3)] |= 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++)
					vram[(y << 4) + i] = 255;
			} else
				vram[(y << 4) + (x1 >> 3)] |= (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			break;
		case ML_WHITE:
			if (x1 >> 3 != x2 >> 3) {
				vram[(y << 4) + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				vram[(y << 4) + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++)
					vram[(y << 4) + i] = 0;
			} else
				vram[(y << 4) + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
			break;
		case ML_XOR:
			if (x1 >> 3 != x2 >> 3) {
				vram[(y << 4) + (x1 >> 3)] ^= 255 >> (x1 & 7);
				vram[(y << 4) + (x2 >> 3)] ^= 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < (x2 >> 3); i++)
					vram[(y << 4) + i] ^= 255;
			} else
				vram[(y << 4) + (x1 >> 3)] ^= (255 >> ((x1 & 7) + 7 - (x2 & 7))) << (7 - (x2 & 7));
			break;
		case ML_CHECKER:
			checker = (y & 1 ? 85 : 170);
			if (x1 >> 3 != x2 >> 3) {
				vram[(y << 4) + (x1 >> 3)] &= 255 << 8 - (x1 & 7);
				vram[(y << 4) + (x2 >> 3)] &= 255 >> 1 + (x2 & 7);
				vram[(y << 4) + (x1 >> 3)] |= checker & 255 >> (x1 & 7);
				vram[(y << 4) + (x2 >> 3)] |= checker & 255 << 7 - (x2 & 7);
				for (i = (x1 >> 3) + 1; i < x2 >> 3; i++)
					vram[(y << 4) + i] = checker;
			} else {
				vram[(y << 4) + (x1 >> 3)] &= (255 << 8 - (x1 & 7)) | (255 >> 1 + (x2 & 7));
				vram[(y << 4) + (x1 >> 3)] |= checker & (255 >> (x1 % 8 + 7 - x2 % 8)) << (7 - (x2 & 7));
			}
			break;
	}
}

void ML_rectangle(int x1, int y1, int x2, int y2, ML_Colour colour) {
	ML_horizontal_line(y1, x1, x2, colour);
	ML_horizontal_line(y2, x1, x2, colour);
	ML_vertical_line(x1 - 1, y1, y2, colour);
	ML_vertical_line(x2 + 1, y1, y2, colour);
}
